// imagelib.c

#include "cmdlib.h"
#include "imagelib.h"


/*
============================================================================

						LBM STUFF

============================================================================
*/


#define FORMID ('F'+('O'<<8)+((long)'R'<<16)+((long)'M'<<24))
#define ILBMID ('I'+('L'<<8)+((long)'B'<<16)+((long)'M'<<24))
#define PBMID  ('P'+('B'<<8)+((long)'M'<<16)+((long)' '<<24))
#define BMHDID ('B'+('M'<<8)+((long)'H'<<16)+((long)'D'<<24))
#define BODYID ('B'+('O'<<8)+((long)'D'<<16)+((long)'Y'<<24))
#define CMAPID ('C'+('M'<<8)+((long)'A'<<16)+((long)'P'<<24))


bmhd_t  bmhd;

long    Align (long l)
{
	if (l&1)
		return l+1;
	return l;
}



/*
================
=
= LBMRLEdecompress
=
= Source must be evenly aligned!
=
================
*/

byte  *LBMRLEDecompress (byte *source,byte *unpacked, int bpwidth)
{
	int     count;
	byte    b,rept;

	count = 0;

	do
	{
		rept = *source++;

		if (rept > 0x80)
		{
			rept = (rept^0xff)+2;
			b = *source++;
			memset(unpacked,b,rept);
			unpacked += rept;
		}
		else if (rept < 0x80)
		{
			rept++;
			memcpy(unpacked,source,rept);
			unpacked += rept;
			source += rept;
		}
		else
			rept = 0;               // rept of 0x80 is NOP

		count += rept;

	} while (count<bpwidth);

	if (count>bpwidth)
		Error ("Decompression exceeded width!\n");


	return source;
}


#define BPLANESIZE      128
byte    bitplanes[9][BPLANESIZE];       // max size 1024 by 9 bit planes


/*
=================
=
= MungeBitPlanes8
=
= This destroys the bit plane data!
=
=================
*/

void MungeBitPlanes8 (int width, byte *dest)
{
	*dest=width;	// shut up the compiler warning
	Error ("MungeBitPlanes8 not rewritten!");
#if 0
asm     les     di,[dest]
asm     mov     si,-1
asm     mov     cx,[width]
mungebyte:
asm     inc     si
asm     mov     dx,8
mungebit:
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*7 +si],1
asm     rcl     al,1
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*6 +si],1
asm     rcl     al,1
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*5 +si],1
asm     rcl     al,1
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*4 +si],1
asm     rcl     al,1
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*3 +si],1
asm     rcl     al,1
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*2 +si],1
asm     rcl     al,1
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*1 +si],1
asm     rcl     al,1
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*0 +si],1
asm     rcl     al,1
asm     stosb
asm     dec     cx
asm     jz      done
asm     dec     dx
asm     jnz     mungebit
asm     jmp     mungebyte

done:
#endif
}


void MungeBitPlanes4 (int width, byte *dest)
{
	*dest=width;	// shut up the compiler warning
	Error ("MungeBitPlanes4 not rewritten!");
#if 0

asm     les     di,[dest]
asm     mov     si,-1
asm     mov     cx,[width]
mungebyte:
asm     inc     si
asm     mov     dx,8
mungebit:
asm     xor     al,al
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*3 +si],1
asm     rcl     al,1
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*2 +si],1
asm     rcl     al,1
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*1 +si],1
asm     rcl     al,1
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*0 +si],1
asm     rcl     al,1
asm     stosb
asm     dec     cx
asm     jz      done
asm     dec     dx
asm     jnz     mungebit
asm     jmp     mungebyte

done:
#endif
}


void MungeBitPlanes2 (int width, byte *dest)
{
	*dest=width;	// shut up the compiler warning
	Error ("MungeBitPlanes2 not rewritten!");
#if 0
asm     les     di,[dest]
asm     mov     si,-1
asm     mov     cx,[width]
mungebyte:
asm     inc     si
asm     mov     dx,8
mungebit:
asm     xor     al,al
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*1 +si],1
asm     rcl     al,1
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*0 +si],1
asm     rcl     al,1
asm     stosb
asm     dec     cx
asm     jz      done
asm     dec     dx
asm     jnz     mungebit
asm     jmp     mungebyte

done:
#endif
}


void MungeBitPlanes1 (int width, byte *dest)
{
	*dest=width;	// shut up the compiler warning
	Error ("MungeBitPlanes1 not rewritten!");
#if 0
asm     les     di,[dest]
asm     mov     si,-1
asm     mov     cx,[width]
mungebyte:
asm     inc     si
asm     mov     dx,8
mungebit:
asm     xor     al,al
asm     shl     [BYTE PTR bitplanes + BPLANESIZE*0 +si],1
asm     rcl     al,1
asm     stosb
asm     dec     cx
asm     jz      done
asm     dec     dx
asm     jnz     mungebit
asm     jmp     mungebyte

done:
#endif
}


/*
=================
=
= LoadLBM
=
=================
*/

void LoadLBM (char *filename, byte **picture, int *width, int *height, byte **palette)
{
	bmhd_t		bmhd;
	byte		*LBMbuffer, *picbuffer, *cmapbuffer;
	int		y,p,planes;
	byte		*LBM_P, *LBMEND_P;
	byte		*pic_p;
	byte		*body_p;
	unsigned        rowsize;

	long    formtype,formlength;
	long    chunktype,chunklength;
	void    (*mungecall) (int, byte *);

//
// load the LBM
//
	LoadFile (filename, (void **)&LBMbuffer);

//
// parse the LBM header
//
	LBM_P = LBMbuffer;
	if ( *(long *)LBMbuffer != LittleLong(FORMID) )
		Error ("No FORM ID at start of file!\n");

	LBM_P += 4;
	formlength = BigLong( *(long *)LBM_P );
	LBM_P += 4;
	LBMEND_P = LBM_P + Align(formlength);

	formtype = LittleLong(*(long *)LBM_P);

	if (formtype != ILBMID && formtype != PBMID)
		Error ("Unrecognized form type: %c%c%c%c\n", formtype&0xff
		,(formtype>>8)&0xff,(formtype>>16)&0xff,(formtype>>24)&0xff);

	LBM_P += 4;

//
// parse chunks
//

	while (LBM_P < LBMEND_P)
	{
		chunktype = *(long *)LBM_P;
		LBM_P += 4;
		chunklength = BigLong(*(long *)LBM_P);
		LBM_P += 4;

		switch (LittleLong(chunktype))
		{
		case BMHDID:
			memcpy (&bmhd,LBM_P,sizeof(bmhd));
			*width = bmhd.w = BigShort(bmhd.w);
			*height = bmhd.h = BigShort(bmhd.h);
			bmhd.x = BigShort(bmhd.x);
			bmhd.y = BigShort(bmhd.y);
			bmhd.pageWidth = BigShort(bmhd.pageWidth);
			bmhd.pageHeight = BigShort(bmhd.pageHeight);
			break;

		case CMAPID:
			cmapbuffer = SafeMalloc (768);
			memset (cmapbuffer, 0, 768);
			memcpy (cmapbuffer, LBM_P, chunklength);
			break;

		case BODYID:
			body_p = LBM_P;

			pic_p = picbuffer = SafeMalloc (bmhd.w*bmhd.h);
			if (formtype == PBMID)
			{
			//
			// unpack PBM
			//
				for (y=0 ; y<bmhd.h ; y++, pic_p += bmhd.w)
				{
					if (bmhd.compression == cm_rle1)
						body_p = LBMRLEDecompress ((byte *)body_p
						, pic_p , bmhd.w);
					else if (bmhd.compression == cm_none)
					{
						memcpy (pic_p,body_p,bmhd.w);
						body_p += Align(bmhd.w);
					}
				}

			}
			else
			{
			//
			// unpack ILBM
			//
				planes = bmhd.nPlanes;
				if (bmhd.masking == ms_mask)
					planes++;
				rowsize = (bmhd.w+15)/16 * 2;
				switch (bmhd.nPlanes)
				{
				case 1:
					mungecall = MungeBitPlanes1;
					break;
				case 2:
					mungecall = MungeBitPlanes2;
					break;
				case 4:
					mungecall = MungeBitPlanes4;
					break;
				case 8:
					mungecall = MungeBitPlanes8;
					break;
				default:
					Error ("Can't munge %i bit planes!\n",bmhd.nPlanes);
				}

				for (y=0 ; y<bmhd.h ; y++, pic_p += bmhd.w)
				{
					for (p=0 ; p<planes ; p++)
						if (bmhd.compression == cm_rle1)
							body_p = LBMRLEDecompress ((byte *)body_p
							, bitplanes[p] , rowsize);
						else if (bmhd.compression == cm_none)
						{
							memcpy (bitplanes[p],body_p,rowsize);
							body_p += rowsize;
						}

					mungecall (bmhd.w , pic_p);
				}
			}
			break;
		}

		LBM_P += Align(chunklength);
	}

	free (LBMbuffer);

	*picture = picbuffer;
	*palette = cmapbuffer;
}

/*
==============
=
= WriteLBMfile
=
==============
*/

void WriteLBMfile (char *filename, byte *data, int width, int height, byte *palette)
{
	byte    *lbm, *lbmptr;
	long    *formlength, *bmhdlength, *cmaplength, *bodylength;
	long    length;
	bmhd_t  basebmhd;

	lbm = lbmptr = SafeMalloc (width*height+1000);

//
// start FORM
//
	*lbmptr++ = 'F';
	*lbmptr++ = 'O';
	*lbmptr++ = 'R';
	*lbmptr++ = 'M';

	formlength = (long*)lbmptr;
	lbmptr+=4;                      // leave space for length

	*lbmptr++ = 'P';
	*lbmptr++ = 'B';
	*lbmptr++ = 'M';
	*lbmptr++ = ' ';

//
// write BMHD
//
	*lbmptr++ = 'B';
	*lbmptr++ = 'M';
	*lbmptr++ = 'H';
	*lbmptr++ = 'D';

	bmhdlength = (long *)lbmptr;
	lbmptr+=4;                      // leave space for length

	memset (&basebmhd,0,sizeof(basebmhd));
	basebmhd.w = BigShort(width);
	basebmhd.h = BigShort(height);
	basebmhd.nPlanes = BigShort(8);
	basebmhd.xAspect = BigShort(5);
	basebmhd.yAspect = BigShort(6);
	basebmhd.pageWidth = BigShort(width);
	basebmhd.pageHeight = BigShort(height);

	memcpy (lbmptr,&basebmhd,sizeof(basebmhd));
	lbmptr += sizeof(basebmhd);

	length = lbmptr-(byte *)bmhdlength-4;
	*bmhdlength = BigLong(length);
	if (length&1)
		*lbmptr++ = 0;          // pad chunk to even offset

//
// write CMAP
//
	*lbmptr++ = 'C';
	*lbmptr++ = 'M';
	*lbmptr++ = 'A';
	*lbmptr++ = 'P';

	cmaplength = (long *)lbmptr;
	lbmptr+=4;                      // leave space for length

	memcpy (lbmptr,palette,768);
	lbmptr += 768;

	length = lbmptr-(byte *)cmaplength-4;
	*cmaplength = BigLong(length);
	if (length&1)
		*lbmptr++ = 0;          // pad chunk to even offset

//
// write BODY
//
	*lbmptr++ = 'B';
	*lbmptr++ = 'O';
	*lbmptr++ = 'D';
	*lbmptr++ = 'Y';

	bodylength = (long *)lbmptr;
	lbmptr+=4;                      // leave space for length

	memcpy (lbmptr,data,width*height);
	lbmptr += width*height;

	length = lbmptr-(byte *)bodylength-4;
	*bodylength = BigLong(length);
	if (length&1)
		*lbmptr++ = 0;          // pad chunk to even offset

//
// done
//
	length = lbmptr-(byte *)formlength-4;
	*formlength = BigLong(length);
	if (length&1)
		*lbmptr++ = 0;          // pad chunk to even offset

//
// write output file
//
	SaveFile (filename, lbm, lbmptr-lbm);
	free (lbm);
}



/*
============================================================================

						PCX STUFF

============================================================================
*/


/*
=================
=
= LoadPCX
=
=================
*/

void LoadPCX (char *filename, byte **picture, int *width, int *height, byte **palette)
{
	byte	*raw;
	pcx_t	*pcx;
	int		x, y;
	int		len;
	int		dataByte, runLength;
	byte	*out, *pix;
	
	//
	// load the file
	//
	len = LoadFile (filename, (void **)&raw);
	
	//
	// parse the PCX file
	//
	pcx = (pcx_t *)raw;
	raw = &pcx->data;
	
	pcx->xmin = LittleShort(pcx->xmin);
	pcx->ymin = LittleShort(pcx->ymin);
	pcx->xmax = LittleShort(pcx->xmax);
	pcx->ymax = LittleShort(pcx->ymax);
	pcx->hres = LittleShort(pcx->hres);
	pcx->vres = LittleShort(pcx->vres);
	pcx->bytes_per_line = LittleShort(pcx->bytes_per_line);
	pcx->palette_type = LittleShort(pcx->palette_type);
	
	if (pcx->manufacturer != 0x0a
		|| pcx->version != 5
		|| pcx->encoding != 1
		|| pcx->bits_per_pixel != 8
		|| pcx->xmax >= 640
		|| pcx->ymax >= 480)
		Error ("Bad pcx file %s", filename);
	
	*palette = SafeMalloc(768);
	memcpy (*palette, (byte *)pcx + len - 768, 768);
	
	*width = pcx->xmax+1;
	*height = pcx->ymax+1;
	
	out = SafeMalloc ((pcx->ymax+1) * (pcx->xmax+1));
	
	*picture = pix = out;
	
	for (y=0 ; y<=pcx->ymax ; y++, pix += pcx->xmax+1)
	{
		for (x=0 ; x<=pcx->xmax ; )
		{
			dataByte = *raw++;
			
			if((dataByte & 0xC0) == 0xC0)
			{
				runLength = dataByte & 0x3F;
				dataByte = *raw++;
			}
			else
				runLength = 1;
			
			// FIXME: this shouldn't happen, but it does.  Are we decoding the file wrong?
			// Truncate runLength so we don't overrun the end of the buffer
			if ((y == pcx->ymax) && (x + runLength > pcx->xmax + 1)) 
				runLength = pcx->xmax - x + 1;
				
			while(runLength-- > 0)
				pix[x++] = dataByte;
		}
		
	}
	
	if (raw - (byte *)pcx > len)
		Error ("PCX file %s was malformed", filename);
	
	free (pcx);
}

/*
==============
=
= WritePCXfile
=
==============
*/

void WritePCXfile (char *filename, byte *data, int width, int height, byte *palette)
{
	int     i, length;
	pcx_t   *pcx;
	byte        *pack;
	
	pcx = SafeMalloc (width*height*2+1000);

	pcx->manufacturer = 0x0a;   // PCX id
	pcx->version = 5;           // 256 color
	pcx->encoding = 1;      // uncompressed
	pcx->bits_per_pixel = 8;        // 256 color
	pcx->xmin = 0;
	pcx->ymin = 0;
	pcx->xmax = LittleShort(width-1);
	pcx->ymax = LittleShort(height-1);
	pcx->hres = LittleShort(width);
	pcx->vres = LittleShort(height);
	memset (pcx->palette,0,sizeof(pcx->palette));
	pcx->color_planes = 1;      // chunky image
	pcx->bytes_per_line = LittleShort(width);
	pcx->palette_type = LittleShort(2);       // not a grey scale
	memset (pcx->filler,0,sizeof(pcx->filler));

//
// pack the image
//
	pack = &pcx->data;

	for (i=0 ; i<width*height ; i++)
		if ( (*data & 0xc0) != 0xc0)
			*pack++ = *data++;
		else
		{
			*pack++ = 0xc1;
			*pack++ = *data++;
		}

//
// write the palette
//
	*pack++ = 0x0c; // palette ID byte
	for (i=0 ; i<768 ; i++)
		*pack++ = *palette++;

//
// write output file
//
	length = pack - (byte *)pcx;
	SaveFile (filename, pcx, length);

	free (pcx);
}

/*
============================================================================

					BMP STUFF

============================================================================
*/

/*
==============
=
= LoadBMP
=
==============
*/

void LoadBMP (char *filename, byte **pic, int *width, int *height, byte **palette)
{
	byte			*out,*data;
	byte			*bmBuffer;
	BITMAPFILEHEADER	*bmFile;
	BITMAPINFO		*bmInfo;
	BITMAPCOREINFO		*bmCore;
	boolean			flipped;
	int			i,structSize;

	LoadFile(filename, (void **)&bmBuffer);
	
	*palette = SafeMalloc(768);
	
	bmFile = (BITMAPFILEHEADER *)bmBuffer;

	if (LittleShort(bmFile->bfType) != 'B' + ('M'<<8))
		Error ("%s is not a bmp file", filename);

	bmFile->bfSize = LittleLong (bmFile->bfSize);
	bmFile->bfOffBits = LittleLong (bmFile->bfOffBits);

	// the size will tell us if it is a
	// bitmapinfo or a bitmapcore
	structSize = LittleLong(*(long *)bmBuffer+sizeof(BITMAPFILEHEADER));
	if (structSize == 40)
	{
		// bitmapinfo
		bmInfo = (BITMAPINFO *)bmFile+sizeof(BITMAPFILEHEADER);
		bmInfo->bmiHeader.biWidth = LittleLong(bmInfo->bmiHeader.biWidth); 
		bmInfo->bmiHeader.biHeight = LittleLong(bmInfo->bmiHeader.biHeight);
		bmInfo->bmiHeader.biPlanes = LittleLong(bmInfo->bmiHeader.biPlanes);
		bmInfo->bmiHeader.biBitCount = LittleLong(bmInfo->bmiHeader.biBitCount);
		
		*width = bmInfo->bmiHeader.biWidth;
		*height = bmInfo->bmiHeader.biHeight;

		for (i = 0 ; i < 256 ; i++)
		{
			(*palette)[i * 3 + 0] = bmInfo->bmiColors[i * 4 + 2];
			(*palette)[i * 3 + 1] = bmInfo->bmiColors[i * 4 + 1];
			(*palette)[i * 3 + 2] = bmInfo->bmiColors[i * 4 + 0];
		}
	} 
	else if (structSize == 12)
	{
		// bitmapcore
		bmCore = (BITMAPCOREINFO *)bmFile+sizeof(BITMAPFILEHEADER);
		bmCore->bmciHeader.bcWidth = LittleShort(bmCore->bmciHeader.bcWidth); 
		bmCore->bmciHeader.bcHeight = LittleShort(bmCore->bmciHeader.bcHeight); 
		bmCore->bmciHeader.bcPlanes = LittleShort(bmCore->bmciHeader.bcPlanes); 
		bmCore->bmciHeader.bcBitCount = LittleShort(bmCore->bmciHeader.bcBitCount);
		
		*width = bmCore->bmciHeader.bcWidth;
		*height = bmCore->bmciHeader.bcHeight;

		for (i = 0 ; i < 256 ; i++)
		{
			(*palette)[i * 3 + 0] = bmCore->bmciColors[i * 3 + 2];
			(*palette)[i * 3 + 1] = bmCore->bmciColors[i * 3 + 1];
			(*palette)[i * 3 + 2] = bmCore->bmciColors[i * 3 + 0];
		}
	} 
	else
		Error ("%s had strange struct size", filename);
	
	if (!(bmCore->bmciHeader.bcPlanes != 1 && bmInfo->bmiHeader.biPlanes != 1))
		Error ("%s was not a single plane image", filename);

	if (!(bmCore->bmciHeader.bcBitCount != 8 && bmInfo->bmiHeader.biBitCount != 8))
		Error ("%s was not an 8 bit image", filename);

	if (*height < 0)
	{
		*height = -(*height);
		flipped = true;
	} 
	else 
		flipped = false;

	*pic = out = SafeMalloc ((*width) * (*height));
	
	data = (byte *)bmBuffer+bmFile->bfOffBits;

	if (flipped)
	{
		for (i = 0; i < *height; i++)
			memcpy((byte *)(out + (*width) * ((*height) - 1 - i)), data, *width);
	} 
	else
		memcpy(out, data, (*width) * (*height));

	free (bmBuffer);
}

/*
==============
=
= WriteBMPfile
=
==============
*/

void WriteBMPfile (char *filename, byte *data, int width, int height, byte *palette)
{
	BITMAPFILEHEADER	*bmFile;
	BITMAPCOREINFO		*bmCore;
	byte			*bmBuffer;
	byte			*picture;
	int			i;
	
	bmBuffer = SafeMalloc(sizeof(BITMAPFILEHEADER)+sizeof(BITMAPCOREINFO)+(width*height));
	bmFile = (BITMAPFILEHEADER *)bmBuffer;
	bmCore = (BITMAPCOREINFO *)bmBuffer+sizeof(BITMAPFILEHEADER);
	
	bmFile->bfType = LittleShort('B' + ('M'<<8));
	bmFile->bfSize = LittleLong(sizeof(BITMAPFILEHEADER)+sizeof(BITMAPCOREINFO)+(width*height));
	bmFile->bfOffBits = LittleLong(sizeof(BITMAPFILEHEADER)+sizeof(BITMAPCOREINFO));
	bmCore->bmciHeader.bcSize = LittleLong(sizeof(BITMAPCOREINFO));
	bmCore->bmciHeader.bcWidth = LittleShort(width);
	bmCore->bmciHeader.bcHeight = LittleShort(height);
	bmCore->bmciHeader.bcPlanes = LittleShort(1);
	bmCore->bmciHeader.bcBitCount = LittleShort(8);
	
	for (i = 0 ; i < 256 ; i++)
	{
		bmCore->bmciColors[i * 3 + 2] = palette[i * 3 + 0];
		bmCore->bmciColors[i * 3 + 1] = palette[i * 3 + 1];
		bmCore->bmciColors[i * 3 + 0] = palette[i * 3 + 2];
	}
	
	picture = (byte *)bmBuffer+bmFile->bfOffBits;
	memcpy(picture, data, width*height);
	
	SaveFile (filename, picture, bmFile->bfSize);
	
	free(bmBuffer);
}

/*
============================================================================

				LOAD/SAVE IMAGE

============================================================================
*/

/*
==============
=
= LoadImage
=
= Will load either an lbm, pcx or bmp, depending on extension.
=
==============
*/

void LoadImage (char *filename, byte **picture, int *width, int *height, byte **palette)
{
	char	ext[128];

	ExtractFileExtension (filename, ext);
	if (!stricmp (ext, "lbm"))
	{
		LoadLBM (filename, picture, width, height, palette);
	}
	else if (!stricmp (ext, "pcx"))
	{
		LoadPCX (filename, picture, width, height, palette);
	}
	else if (!stricmp (ext, "bmp"))
	{
		LoadBMP (filename, picture, width, height, palette);
	}
	else
		Error ("%s doesn't have a known image extension", filename);
}


/*
==============
=
= WriteImage
=
= Will save either an lbm, pcx or bmp, depending on extension.
=
==============
*/

void WriteImage (char *filename, byte *data, int width, int height, byte *palette)
{
	char	ext[128];

	ExtractFileExtension (filename, ext);
	if (!stricmp (ext, "lbm"))
	{
		WriteLBMfile (filename, data, width, height, palette);
	}
	else if (!stricmp (ext, "pcx"))
	{
		WritePCXfile (filename, data, width, height, palette);
	}
	else if (!stricmp (ext, "bmp"))
	{
		WriteBMPfile (filename, data, width, height, palette);
	}
	else
		Error ("%s doesn't have a known image extension", filename);
}

